;; Tree operations (bst)

(define-struct node (soc l r p))

;; Original tree
(define lina (make-node 99 empty empty empty))
(define ennio (make-node 24 empty empty empty))
(define elvira (make-node 10 empty empty empty))
(define tina (make-node 95 empty lina empty))
(define paride (make-node 77 empty empty empty))
(define teresa (make-node 15 elvira ennio empty))
(define patrizia (make-node 89 paride tina empty))
(define giuseppe (make-node 29 teresa empty empty))
(define luca (make-node 63 giuseppe patrizia empty))

(set-node-p! giuseppe luca)
(set-node-p! patrizia luca)
(set-node-p! teresa giuseppe)
(set-node-p! paride patrizia)
(set-node-p! tina patrizia)
(set-node-p! elvira teresa)
(set-node-p! ennio teresa)
(set-node-p! lina tina)

;; Nodes for examples
(define celso (make-node 35 empty empty empty))
(define ottorino (make-node 92 empty empty empty))

;;----------------------------
;; Aux functions
;;----------------------------

(define (tree-minimum atree)
  (cond
    [(empty? atree) empty]
    [else
     (cond
       [(empty? (node-l atree)) atree]
       [else (tree-minimum (node-l atree))])]))

(define (tree-successor atree)
  (cond
    [(empty? atree) empty]
    [else
     (cond
       [(empty? (node-r atree))
        (cond
          [(= (node-soc atree) (node-soc (node-r (node-p atree)))) (tree-successor (node-p atree))]
          [else (node-p atree)])]
       [else (tree-minimum (node-r atree))])]))


(define (extract-successor atree)
  (cond
    ;; se il succ e' a sinistra
    [(= (node-soc (tree-successor atree)) (node-soc (node-l (node-p (tree-successor atree)))))
     (cond
       ;; se non ha figlio vuotiamo il sinistro di suo padre
       [(empty? (node-r (tree-successor atree)))
        (begin
          (tree-successor atree)
          (set-node-l! (node-p (tree-successor atree)) empty))]
       ;; se ha figlio (destro) lo linkiamo al sinistro di suo padre
       [else
        (begin
          (tree-successor atree)
          (set-node-p! (node-r (tree-successor atree)) (node-p (tree-successor atree)))
          (set-node-l! (node-p (tree-successor atree)) (node-r (tree-successor atree))))])]
    
    ;; se il succ e' a destra
    [else
     (cond
       ;; se non ha figlio vuotiamo il destro di suo padre
       [(empty? (node-r (tree-successor atree)))
        (begin
          (tree-successor atree)
          (set-node-r! (node-p (tree-successor atree)) empty))]
       ;; se ha figlio (destro) lo linkiamo al destro di suo padre
       [else
        (begin
          (tree-successor atree)
          (set-node-p! (node-r (tree-successor atree)) (node-p (tree-successor atree)))
          (set-node-r! (node-p (tree-successor atree)) (node-r (tree-successor atree))))])]))


;;----------------------------
;; Search
;;----------------------------

(define (tree-search anumb atree)
  (cond
    [(empty? atree) false]
    [else
     (cond
       [(= anumb (node-soc atree)) true]
       [(< anumb (node-soc atree)) (tree-search anumb (node-l atree))]
       [(> anumb (node-soc atree)) (tree-search anumb (node-r atree))]
       [else false])]))

;;----------------------------
;; Delete
;;----------------------------

(define (tree-delete anode atree)
  (cond
    [(empty? atree) false]
    [(tree-search (node-soc anode) atree)
     (cond
       
       ;; Son-l and son-r are empty
       [(and (empty? (node-l anode)) (empty? (node-r anode)))
        (cond
          [(= (node-soc anode) (node-soc (node-l (node-p anode)))) (set-node-l! (node-p anode) empty)]
          [else (set-node-r! (node-p anode) empty)])]
       
       ;; Son-l is empty
       [(empty? (node-l anode))
        (cond
          [(= (node-soc anode) (node-soc (node-l (node-p anode))))
           (begin (set-node-l! (node-p anode) (node-r anode)) (set-node-p! (node-r anode) (node-p anode)))]
          [else 
           (begin (set-node-r! (node-p anode) (node-r anode)) (set-node-p! (node-r anode) (node-p anode)))])]
       
       ;; Son-r is empty
       [(empty? (node-r anode))
        (cond
          [(= (node-soc anode) (node-soc (node-l (node-p anode))))
           (begin (set-node-l! (node-p anode) (node-l anode)) (set-node-p! (node-l anode) (node-p anode)))]
          [else
           (begin (set-node-r! (node-p anode) (node-l anode)) (set-node-p! (node-l anode) (node-p anode)))])]
       
       ;; Son-l and son-r are not empty
       [else
        (cond
          ;; se stiamo a sinistra
          [(= (node-soc anode) (node-soc (node-l (node-p anode))))
           (begin
             (set-node-l! (node-p anode) (extract-successor anode))
             (set-node-p! (node-l anode) (node-l (node-p anode)))
             (set-node-p! (node-r anode) (node-l (node-p anode)))
             (set-node-p! (node-l (node-p anode)) (node-p anode))
             (set-node-l! (node-l (node-p anode)) (node-l anode))
             (set-node-r! (node-l (node-p anode)) (node-r anode)))]
          ;; se stiamo a destra
          [else
           (begin
             (set-node-r! (node-p anode) (extract-successor anode))
             (set-node-p! (node-l anode) (node-r (node-p anode)))
             (set-node-p! (node-r anode) (node-r (node-p anode)))
             (set-node-p! (node-r (node-p anode)) (node-p anode))
             (set-node-l! (node-r (node-p anode)) (node-l anode))
             (set-node-r! (node-r (node-p anode)) (node-r anode)))])])]
    [else false]))

;;----------------------------
;; Insert
;;----------------------------

(define (tree-insert anode atree)
  (cond
    [(empty? atree) (node-soc anode)]
    [else
     (cond
       [(< (node-soc anode) (node-soc atree))
        (cond
          [(empty? (node-l atree)) (node-soc atree)]
          [else (tree-insert anode (node-l atree))])]
       [(> (node-soc anode) (node-soc atree))
        (cond
          [(empty? (node-r atree)) (node-soc atree)]
          [else (tree-insert anode (node-r atree))])]
       [else false])]))